from django.shortcuts import render
from django.urls import reverse


def generic(request):
    context = {
        "path": request.get_full_path(),
        "user": request.user,
        "schema": request.tenant.schema_name,
        "routing": request.tenant.routing,
        "admin_url": reverse("admin:index"),
    }
    return render(request, "index.html", context)



import pytest
from django.core import management


@pytest.fixture(autouse=True)
def _setup(tenant1, tenant2, tenant3, TenantModel, DomainModel):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    if DomainModel:
        for tenant in [tenant1, tenant2, tenant3]:
            DomainModel.objects.create(
                tenant=tenant, domain=f"{tenant.schema_name}.localhost", is_primary=True
            )


def split_output(buffer):
    buffer.seek(0)
    return set(buffer.read().strip().splitlines())


def test_all_schemas(DomainModel, stdout):
    management.call_command("whowill", all_schemas=True, stdout=stdout)

    expected_dynamic = (
        {"tenant1.localhost", "tenant2.localhost", "tenant3.localhost"}
        if DomainModel
        else {"tenant1", "tenant2", "tenant3"}
    )

    assert (
        split_output(stdout)
        == {"public", "sample", "localhost", "blog.localhost"} | expected_dynamic
    )


def test_static_schemas(stdout):
    management.call_command("whowill", static_schemas=True, stdout=stdout)

    assert split_output(stdout) == {"public", "sample", "localhost", "blog.localhost"}


def test_tenant_like_schemas(DomainModel, stdout):
    management.call_command("whowill", tenant_schemas=True, stdout=stdout)

    expected_dynamic = (
        {"tenant1.localhost", "tenant2.localhost", "tenant3.localhost"}
        if DomainModel
        else {"tenant1", "tenant2", "tenant3"}
    )

    assert split_output(stdout) == {"sample"} | expected_dynamic


def test_dynamic_schemas(DomainModel, stdout):
    management.call_command("whowill", dynamic_schemas=True, stdout=stdout)

    expected_dynamic = (
        {"tenant1.localhost", "tenant2.localhost", "tenant3.localhost"}
        if DomainModel
        else {"tenant1", "tenant2", "tenant3"}
    )

    assert split_output(stdout) == expected_dynamic


def test_specific_schemas(DomainModel, stdout):
    management.call_command("whowill", schemas=["www", "blog", "tenant1"], stdout=stdout)

    expected_dynamic = {"tenant1.localhost"} if DomainModel else {"tenant1"}

    assert split_output(stdout) == {"localhost", "blog.localhost"} | expected_dynamic


# Same test cases as before, but excluding one


def test_all_schemas_minus_one(DomainModel, stdout):
    management.call_command("whowill", all_schemas=True, excluded_schemas=["blog"], stdout=stdout)

    expected_dynamic = (
        {
            "tenant1.localhost",
            "tenant2.localhost",
            "tenant3.localhost",
        }
        if DomainModel
        else {"tenant1", "tenant2", "tenant3"}
    )

    assert split_output(stdout) == {"public", "sample", "localhost"} | expected_dynamic


def test_static_schemas_minus_one(stdout):
    management.call_command(
        "whowill", static_schemas=True, excluded_schemas=["sample"], stdout=stdout
    )

    assert split_output(stdout) == {"public", "localhost", "blog.localhost"}


def test_tenant_like_schemas_minus_one(DomainModel, stdout):
    management.call_command(
        "whowill", tenant_schemas=True, excluded_schemas=["tenant1"], stdout=stdout
    )

    expected_dynamic = (
        {"tenant2.localhost", "tenant3.localhost"} if DomainModel else {"tenant2", "tenant3"}
    )

    assert split_output(stdout) == {"sample"} | expected_dynamic


def test_dynamic_schemas_minus_one(DomainModel, stdout):
    management.call_command(
        "whowill", dynamic_schemas=True, excluded_schemas=["public"], stdout=stdout
    )

    expected_dynamic = (
        {
            "tenant1.localhost",
            "tenant2.localhost",
            "tenant3.localhost",
        }
        if DomainModel
        else {
            "tenant1",
            "tenant2",
            "tenant3",
        }
    )

    assert split_output(stdout) == expected_dynamic


def test_specific_schemas_minus_one(DomainModel, stdout):
    management.call_command(
        "whowill",
        schemas=["www", "blog", "tenant1"],
        excluded_schemas=["www"],
        stdout=stdout,
    )

    expected_dynamic = {"tenant1.localhost"} if DomainModel else {"tenant1"}

    assert split_output(stdout) == {"blog.localhost"} | expected_dynamic



from unittest.mock import patch

import pytest
from django.core import management
from django.core.management.base import CommandError

from django_pgschemas.management.commands import CommandScope
from django_pgschemas.management.commands.whowill import Command as WhoWillCommand


@pytest.fixture(autouse=True)
def _setup(tenant1, tenant2, TenantModel, DomainModel):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    if DomainModel:
        DomainModel.objects.create(
            tenant=tenant1,
            domain="tenant1.localhost",
            is_primary=True,
        )
        DomainModel.objects.create(
            tenant=tenant1,
            domain="everyone.localhost",
            folder="tenant1",
            is_primary=False,
        )
        DomainModel.objects.create(
            tenant=tenant2,
            domain="tenant2.localhost",
            is_primary=True,
        )
        DomainModel.objects.create(
            tenant=tenant2,
            domain="everyone.localhost",
            folder="tenant2",
            is_primary=False,
        )


def test_no_schema_provided():
    command = WhoWillCommand()

    with pytest.raises(CommandError) as ctx:
        management.call_command(command, interactive=False, verbosity=0)

    assert str(ctx.value) == "No schema provided"


def test_no_all_schemas_allowed():
    command = WhoWillCommand()
    command.allow_wildcards = False

    with pytest.raises(TypeError):
        management.call_command(command, all_schemas=True, verbosity=0)


def test_no_static_schemas_allowed():
    command = WhoWillCommand()
    command.scope = CommandScope.DYNAMIC

    with pytest.raises(CommandError) as ctx:
        management.call_command(command, static_schemas=True, verbosity=0)

    assert str(ctx.value) == "Including static schemas is NOT allowed"

    command = WhoWillCommand()
    command.allow_wildcards = False

    with pytest.raises(TypeError):
        management.call_command(command, static_schemas=True, verbosity=0)


def test_no_dynamic_schemas_allowed():
    command = WhoWillCommand()
    command.scope = CommandScope.STATIC

    with pytest.raises(CommandError) as ctx:
        management.call_command(command, dynamic_schemas=True, verbosity=0)

    assert str(ctx.value) == "Including dynamic schemas is NOT allowed"

    command = WhoWillCommand()
    command.allow_wildcards = False

    with pytest.raises(TypeError):
        management.call_command(command, dynamic_schemas=True, verbosity=0)


def test_no_tenant_like_schemas_allowed():
    command = WhoWillCommand()
    command.scope = CommandScope.STATIC

    with pytest.raises(CommandError) as ctx:
        management.call_command(command, tenant_schemas=True, verbosity=0)

    assert str(ctx.value) == "Including tenant-like schemas is NOT allowed"

    command = WhoWillCommand()
    command.allow_wildcards = False

    with pytest.raises(TypeError):
        management.call_command(command, tenant_schemas=True, verbosity=0)


def test_nonexisting_schema():
    with pytest.raises(CommandError) as ctx:
        management.call_command("whowill", schemas=["unknown"], verbosity=0)

    assert str(ctx.value) == "No schema found for 'unknown'"


def test_ambiguous_schema(DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    with pytest.raises(CommandError) as ctx:
        management.call_command("whowill", schemas=["tenant"], verbosity=0)

    assert (
        str(ctx.value)
        == "More than one tenant found for schema 'tenant' by domain, please, narrow down the filter"
    )


def test_specific_schemas():
    command = WhoWillCommand()
    command.specific_schemas = ["blog"]

    with pytest.raises(CommandError) as ctx:
        management.call_command(command, schemas=["www"], verbosity=0)

    assert str(ctx.value) == "This command can only run in ['blog']"


def test_nonexisting_schema_excluded():
    with pytest.raises(CommandError) as ctx:
        management.call_command(
            "whowill", all_schemas=True, excluded_schemas=["unknown"], verbosity=0
        )

    assert str(ctx.value) == "No schema found for 'unknown' (excluded)"


def test_ambiguous_schema_excluded(DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    with pytest.raises(CommandError) as ctx:
        management.call_command(
            "whowill", all_schemas=True, excluded_schemas=["tenant"], verbosity=0
        )

    assert (
        str(ctx.value)
        == "More than one tenant found for schema 'tenant' by domain (excluded), please, narrow down the filter"
    )


def test_existing_schema_excluded_ok():
    management.call_command("whowill", all_schemas=True, excluded_schemas=["tenant1"], verbosity=0)


def test_interactive_ok():
    def patched_input(*args, **kwargs):
        return "blog"

    with patch("builtins.input", patched_input):
        management.call_command("whowill", schemas=[], verbosity=0)


def test_interactive_nonexisting():
    def patched_input(*args, **kwargs):
        return "unknown"

    with patch("builtins.input", patched_input):
        with pytest.raises(CommandError) as ctx:
            management.call_command("whowill", schemas=[], verbosity=0)

        assert str(ctx.value) == "No schema found for 'unknown'"


def test_mixed_ok(DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    management.call_command("whowill", all_schemas=True, verbosity=0)
    management.call_command("whowill", static_schemas=True, verbosity=0)
    management.call_command("whowill", dynamic_schemas=True, verbosity=0)
    management.call_command("whowill", tenant_schemas=True, verbosity=0)
    management.call_command("whowill", schemas=["public", "sample"], verbosity=0)
    management.call_command(
        "whowill",
        all_schemas=True,
        static_schemas=True,
        dynamic_schemas=True,
        tenant_schemas=True,
        schemas=["public", "sample"],
        verbosity=0,
    )
    management.call_command(
        "whowill", all_schemas=True, excluded_schemas=["public", "sample"], verbosity=0
    )
    management.call_command("whowill", schemas=["everyone.localhost/tenant1"], verbosity=0)
    management.call_command("whowill", schemas=["tenant1"], verbosity=0)
    management.call_command(
        "whowill",
        all_schemas=True,
        excluded_schemas=["everyone.localhost/tenant1"],
        verbosity=0,
    )
    management.call_command("whowill", all_schemas=True, excluded_schemas=["tenant1"], verbosity=0)



import pytest
from django.core import management


@pytest.fixture(autouse=True)
def _setup(TenantModel, DomainModel, db):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    tenants = []

    for i in range(10, 20):
        tenant = TenantModel(schema_name=f"tenant{i + 1}")
        tenant.save(verbosity=0)
        if DomainModel:
            DomainModel.objects.create(
                tenant=tenant, domain=f"tenant{i + 1}.localhost", is_primary=True
            )

        tenants.append(tenant)

    yield

    for tenant in tenants:
        tenant.delete(force_drop=True)


def test_all_schemas_in_sequential():
    # If there are no errors, then this test passed
    management.call_command("migrate", all_schemas=True, parallel=False, verbosity=0)


def test_all_schemas_in_parallel():
    # If there are no errors, then this test passed
    management.call_command("migrate", all_schemas=True, parallel=True, verbosity=0)



# Generated by Django 2.1.4 on 2019-01-13 03:07

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("shared_public", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="TenantData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "catalog",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tenant_objects",
                        to="shared_public.Catalog",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="tenant_objects",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        )
    ]



from django.contrib import admin
from django.contrib.auth.decorators import login_required
from django.urls import path

from sandbox.views import generic

urlpatterns = [
    path("", generic, name="tenant-home"),
    path("profile/", generic, name="profile"),
    path("profile/advanced/", login_required(generic), name="advanced-profile"),
    path("login/", generic, name="login"),
    path("admin/", admin.site.urls),
]



import warnings
from unittest.mock import patch

import pytest
from django.apps import apps
from django.core import management
from django.core.management.base import CommandError
from django.db import models
from django.db.utils import ProgrammingError

from django_pgschemas.checks import check_schema_names
from django_pgschemas.models import TenantModel as BaseTenantModel
from django_pgschemas.utils import get_tenant_model


def patched_get_tenant_model(*args, **kwargs):
    if RealTenantModel := get_tenant_model():

        class TenantModel(BaseTenantModel):
            dummy = models.TextField()

            class Meta:
                app_label = RealTenantModel._meta.app_label

        return TenantModel

    return None


@pytest.mark.bug
def test_database_checks_with_zero_migrations(transactional_db):
    """
    Provoke a handled ProgrammingError by migrating models from empty database.
    """
    management.call_command("migrate", "shared_public", "zero", verbosity=0)
    # The goal is that the next line doesn't raise ProgrammingError
    check_schema_names(apps.get_app_config("django_pgschemas"))
    management.call_command("migrate", verbosity=0)


@pytest.mark.bug
@patch("django_pgschemas.management.commands.get_tenant_model", patched_get_tenant_model)
def test_whowill_with_pending_migrations(TenantModel, db):
    """
    Provoke a handled ProgrammingError by running tenant command with pending model changes.
    """
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")  # Avoid warnings about model being registered twice

        with pytest.raises(CommandError) as ctx:
            management.call_command("whowill", all_schemas=True, verbosity=0)

        assert str(ctx.value) == (
            "Error while attempting to retrieve dynamic schemas. "
            "Perhaps you need to migrate the 'public' schema first?"
        )


@pytest.mark.bug
def test_migrate_with_exclusions(TenantModel, db):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    # We first unapply a migration with fake so we can reapply it without fake
    # This should work without errors
    management.call_command(
        "migrate", "app_tenants", "0001_initial", fake=True, schemas=["tenant1"], verbosity=0
    )

    # We then migrate on all schemas except for tenant1, THIS IS THE CASE WE WANT TO TEST
    # This should work without errors
    management.call_command("migrate", all_schemas=True, excluded_schemas=["tenant1"], verbosity=0)

    # If we try to global migrate now, we should get a ProgrammingError
    with pytest.raises(ProgrammingError):
        management.call_command("migrate", all_schemas=True, verbosity=0)
    # We finally apply the migration again with fake
    # This should work without errors

    management.call_command("migrate", fake=True, all_schemas=True, verbosity=0)



# Generated by Django 2.1.4 on 2019-01-13 04:41

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="MainData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
            ],
        ),
    ]



import pytest
from django.core.exceptions import ValidationError
from django.db import connection
from django.db.utils import DatabaseError

from django_pgschemas import schema, utils

VALID_IDENTIFIERS = ["___", "a_a0", "_a0_", "a" * 63]
INVALID_IDENTIFIERS = ["", " ", "^", ".", "&", "{", "(", "@", "!", "a" * 64]
VALID_SCHEMA_NAMES = ["a_pg", "w_pg_a", "_pg_awa", "pgwa"] + VALID_IDENTIFIERS
INVALID_SCHEMA_NAMES = ["pg_a", "pg_"] + INVALID_IDENTIFIERS


def test_get_tenant_model(tenants_settings):
    TenantModel = utils.get_tenant_model()

    if "default" in tenants_settings:
        assert TenantModel is not None
        assert TenantModel._meta.model_name == "tenant"
    else:
        assert TenantModel is None


def test_get_domain_model(tenants_settings):
    DomainModel = utils.get_domain_model()

    if "default" in tenants_settings and "DOMAIN_MODEL" in tenants_settings["default"]:
        assert DomainModel is not None
        assert DomainModel._meta.model_name == "domain"
    else:
        assert DomainModel is None


@pytest.mark.parametrize("has_value", [True, False])
def test_get_tenant_database_alias(settings, has_value):
    if has_value:
        settings.PGSCHEMAS_TENANT_DB_ALIAS = "something"
        assert utils.get_tenant_database_alias() == "something"
    else:
        assert utils.get_tenant_database_alias() == "default"


@pytest.mark.parametrize("has_value", [True, False])
def test_get_limit_set_calls(settings, has_value):
    if has_value:
        settings.PGSCHEMAS_LIMIT_SET_CALLS = True
        assert utils.get_limit_set_calls()
    else:
        assert not utils.get_limit_set_calls()


def test_get_clone_reference(tenants_settings):
    clone_reference = utils.get_clone_reference()

    if "default" in tenants_settings:
        assert clone_reference == "sample"
    else:
        assert clone_reference is None


@pytest.mark.parametrize(
    "identifier, is_valid",
    [(identifier, True) for identifier in VALID_IDENTIFIERS]
    + [(identifier, False) for identifier in INVALID_IDENTIFIERS],
)
def test_is_valid_identifier(identifier, is_valid):
    assert utils.is_valid_identifier(identifier) == is_valid


@pytest.mark.parametrize(
    "name, is_valid",
    [(name, True) for name in VALID_SCHEMA_NAMES]
    + [(name, False) for name in INVALID_SCHEMA_NAMES],
)
def test_is_valid_schema_name(name, is_valid):
    assert utils.is_valid_schema_name(name) == is_valid


@pytest.mark.parametrize(
    "name, is_valid",
    [(name, True) for name in VALID_SCHEMA_NAMES]
    + [(name, False) for name in INVALID_SCHEMA_NAMES],
)
def test_check_schema_name(name, is_valid):
    if is_valid:
        utils.check_schema_name(name)
    else:
        with pytest.raises(ValidationError):
            utils.check_schema_name(name)


def test_run_in_public_schema(db):
    @utils.run_in_public_schema
    def inner():
        with connection.cursor() as cursor:
            cursor.execute("SHOW search_path")
            assert cursor.fetchone() == ("public",)

    with schema.Schema.create(schema_name="test"):
        inner()
        with connection.cursor() as cursor:
            cursor.execute("SHOW search_path")
            cursor.fetchone() == ("test, public",)


def test_schema_exists(db):
    assert utils.schema_exists("public")
    assert utils.schema_exists("www")
    assert utils.schema_exists("blog")
    assert not utils.schema_exists("default")


def test_dynamic_models_exist(tenants_settings, db):
    if "default" in tenants_settings:
        assert utils.dynamic_models_exist()
    else:
        assert not utils.dynamic_models_exist()

    utils.drop_schema("public")

    assert not utils.dynamic_models_exist()


def test_create_drop_schema(db):
    assert not utils.create_schema("public", check_if_exists=True)  # Schema existed already
    assert utils.schema_exists("public")  # Schema exists
    assert utils.drop_schema("public")  # Schema was dropped
    assert not utils.drop_schema("public")  # Schema no longer exists
    assert not utils.schema_exists("public")  # Schema doesn't exist
    assert utils.create_schema("public", sync_schema=False)  # Schema was created
    assert utils.schema_exists("public")  # Schema exists


def test_clone_schema(db):
    utils._create_clone_schema_function()

    assert not utils.schema_exists("sample2")  # Schema doesn't exist previously

    utils.clone_schema("sample", "sample2", dry_run=True)  # Dry run

    assert not utils.schema_exists("sample2")  # Schema won't exist, dry run

    utils.clone_schema("sample", "sample2")  # Real run, schema was cloned

    assert utils.schema_exists("sample2")  # Schema exists

    with pytest.raises(DatabaseError):
        utils.clone_schema("sample", "sample2")  # Schema already exists, error

    assert utils.schema_exists("sample2")  # Schema still exists


def test_create_or_clone_schema(db):
    assert not utils.create_or_clone_schema("sample")  # Schema existed


@pytest.mark.parametrize(
    "path, expected",
    [
        ("", ""),
        ("www", "www"),
        ("www.", ""),
        ("www.test.com", "test.com"),
        ("www.test.com/complex/path", "test.com/complex/path"),
        ("1www.test.com", "1www.test.com"),
    ],
)
def test_remove_www(path, expected):
    actual = utils.remove_www(path)

    assert actual == expected



import pytest
from django.apps import apps
from django.test import Client

from django_pgschemas.routing.info import DomainInfo
from django_pgschemas.schema import Schema


@pytest.fixture
def UserModel():
    return apps.get_model("shared_common.User")


@pytest.fixture(autouse=True)
def _setup(UserModel, db):
    with Schema.create("www"):
        UserModel.objects.create(email="user_www@localhost", display_name="Admin")

    with Schema.create("blog"):
        UserModel.objects.create(email="user_blog@localhost", display_name="Admin")


@pytest.fixture
def _setup_dynamic(tenant1, UserModel, DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    DomainModel.objects.create(tenant=tenant1, domain="tenant1.localhost", is_primary=True)
    DomainModel.objects.create(tenant=tenant1, domain="everyone.localhost", folder="tenant1")

    with tenant1:
        yield UserModel.objects.create(email="user1@localhost", display_name="Admin")


@pytest.mark.parametrize(
    "url, expected_status",
    [
        ("/", 200),
        ("/register/", 200),
        ("/admin/", 302),
        ("/non-existing/", 404),
    ],
)
def test_views_www(url, expected_status):
    client = Client(headers={"host": "localhost"})

    response = client.get(url)

    assert response.status_code == expected_status


@pytest.mark.parametrize(
    "url, expected_status",
    [
        ("/", 200),
        ("/entries/", 200),
        ("/admin/", 302),
        ("/non-existing/", 404),
    ],
)
def test_views_blog(url, expected_status):
    client = Client(headers={"host": "blog.localhost"})

    response = client.get(url)

    assert response.status_code == expected_status


@pytest.mark.parametrize(
    "url, expected_status",
    [
        ("/", 200),
        ("/profile/", 200),
        ("/profile/advanced/", 302),
        ("/login/", 200),
        ("/admin/", 302),
        ("/non-existing/", 404),
    ],
)
def test_tenants_domain(url, expected_status, _setup_dynamic):
    user = _setup_dynamic
    client = Client(headers={"host": "tenant1.localhost"})

    response = client.get(url)

    assert response.status_code == expected_status

    if expected_status == "200":
        assert response.context == {
            "path": url,
            "user": user,
            "schema": "tenant1",
            "routing": DomainInfo(domain="tenant1.localhost", folder=""),
            "admin_url": "/admin/",
        }


@pytest.mark.parametrize(
    "url, expected_status",
    [
        ("/tenant1/", 200),
        ("/tenant1/profile/", 200),
        ("/tenant1/profile/advanced/", 302),
        ("/tenant1/login/", 200),
        ("/tenant1/admin/", 302),
        ("/tenant1/non-existing/", 404),
    ],
)
def test_tenants_folder(url, expected_status, _setup_dynamic):
    user = _setup_dynamic
    client = Client(headers={"host": "everyone.localhost"})

    response = client.get(url)

    assert response.status_code == expected_status

    if expected_status == "200":
        assert response.context == {
            "path": url,
            "user": user,
            "schema": "tenant1",
            "routing": DomainInfo(domain="everyone.localhost", folder="tenant1"),
            "admin_url": "/tenant1/admin/",
        }



import pytest
from django.template import Context, Template

from django_pgschemas.routing.info import DomainInfo, HeadersInfo, SessionInfo
from django_pgschemas.schema import (
    Schema,
    deactivate,
    get_current_schema,
    get_default_schema,
    override,
    shallow_equal,
)


@pytest.mark.parametrize(
    "schema1, schema2, equals",
    [
        (
            get_default_schema(),
            get_default_schema(),
            True,
        ),
        (
            Schema.create("test1"),
            Schema.create("test1"),
            True,
        ),
        (
            Schema.create("test1"),
            Schema.create("test2"),
            False,
        ),
        (
            Schema.create("test1"),
            Schema.create("test1", SessionInfo("ref1")),
            False,
        ),
        (
            Schema.create("test1", HeadersInfo("ref1")),
            Schema.create("test1", SessionInfo("ref1")),
            False,
        ),
        (
            Schema.create("test1", HeadersInfo("ref1")),
            Schema.create("test1", HeadersInfo("ref1")),
            True,
        ),
        (
            Schema.create("test1", DomainInfo("domain1")),
            Schema.create("test1", DomainInfo("domain1", "folder")),
            False,
        ),
        (
            Schema.create("test1", DomainInfo("domain1", "folder")),
            Schema.create("test1", DomainInfo("domain1", "folder")),
            True,
        ),
    ],
)
def test_shallow_equal(schema1, schema2, equals):
    assert shallow_equal(schema1, schema2) == equals


def test_nested_override():
    deactivate()

    schema1 = Schema.create(schema_name="schema_1")
    schema2 = Schema.create(schema_name="schema_2")

    assert get_current_schema().schema_name == get_default_schema().schema_name

    with override(schema1):
        assert get_current_schema().schema_name == schema1.schema_name

        with override(schema2):
            assert get_current_schema().schema_name == schema2.schema_name

            with override(schema1):
                assert get_current_schema().schema_name == schema1.schema_name

            assert get_current_schema().schema_name == schema2.schema_name

        assert get_current_schema().schema_name == schema1.schema_name

    assert get_current_schema().schema_name == get_default_schema().schema_name


def test_nested_class_override():
    deactivate()

    schema1 = Schema.create(schema_name="schema_1")
    schema2 = Schema.create(schema_name="schema_2")

    assert get_current_schema().schema_name == get_default_schema().schema_name

    with schema1:
        assert get_current_schema().schema_name == schema1.schema_name

        with schema2:
            assert get_current_schema().schema_name == schema2.schema_name

            with schema1:
                assert get_current_schema().schema_name == schema1.schema_name

            assert get_current_schema().schema_name == schema2.schema_name

        assert get_current_schema().schema_name == schema1.schema_name

    assert get_current_schema().schema_name == get_default_schema().schema_name


def test_schema_is_template_renderable():
    schema = Schema.create(schema_name="template_schema")

    context = Context({"schema": schema})
    template = Template("{{ schema.schema_name }}")

    rendered = template.render(context)

    assert rendered == "template_schema"



from django_pgschemas.contrib.cache import make_key, reverse_key
from django_pgschemas.schema import Schema


def test_make_key_with_dynamic_tenant(tenant1):
    with tenant1:
        key = make_key(key="foo", key_prefix="", version=1)

    chunks = key.split(":")

    assert len(chunks) == 4
    assert str(tenant1.schema_name) == chunks[0]


def test_make_key_with_static_tenant():
    with Schema.create(schema_name="www"):
        key = make_key(key="foo", key_prefix="", version=1)

    chunks = key.split(":")

    assert len(chunks) == 4
    assert "www" == chunks[0]


def test_reverse_key(tenant1):
    key = "some-key"

    with tenant1:
        assert key == reverse_key(make_key(key=key, key_prefix="", version=1))



from io import StringIO

import pytest


@pytest.fixture(scope="session", autouse=True)
def setup(django_db_setup, django_db_blocker):
    from sandbox.shared_public.models import Tenant

    with django_db_blocker.unblock():
        Tenant.objects.get_or_create(schema_name="tenant1")
        Tenant.objects.get_or_create(schema_name="tenant2")
        Tenant.objects.get_or_create(schema_name="tenant3")


@pytest.fixture(autouse=True, params=["static-only", "tenants-no-domains", "tenants-and-domains"])
def tenants_settings(request, settings):
    from copy import deepcopy

    current = deepcopy(settings.TENANTS)

    if request.param == "static-only":
        del settings.TENANTS["default"]

    if request.param == "tenants-no-domains":
        del settings.TENANTS["default"]["DOMAIN_MODEL"]

    yield settings.TENANTS

    settings.TENANTS.clear()
    settings.TENANTS.update(current)


@pytest.fixture
def TenantModel():
    from django_pgschemas.utils import get_tenant_model

    return get_tenant_model()


@pytest.fixture
def DomainModel():
    from django_pgschemas.utils import get_domain_model

    return get_domain_model()


@pytest.fixture
def tenant1(db):
    from sandbox.shared_public.models import Tenant

    return Tenant.objects.get(schema_name="tenant1")


@pytest.fixture
def tenant2(db):
    from sandbox.shared_public.models import Tenant

    return Tenant.objects.get(schema_name="tenant2")


@pytest.fixture
def tenant3(db):
    from sandbox.shared_public.models import Tenant

    return Tenant.objects.get(schema_name="tenant3")


@pytest.fixture
def stdout():
    with StringIO() as buffer:
        yield buffer



# Generated by Django 2.1.4 on 2019-01-13 03:07

import django.db.models.deletion
from django.db import migrations, models

import django_pgschemas.schema
import django_pgschemas.utils


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Domain",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("domain", models.CharField(db_index=True, max_length=253)),
                ("folder", models.SlugField(blank=True, max_length=253)),
                ("is_primary", models.BooleanField(default=True)),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="Catalog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                )
            ],
        ),
        migrations.CreateModel(
            name="Tenant",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "schema_name",
                    models.CharField(
                        max_length=63,
                        unique=True,
                        validators=[django_pgschemas.utils.check_schema_name],
                    ),
                ),
            ],
            options={"abstract": False},
            bases=(django_pgschemas.schema.Schema, models.Model),
        ),
        migrations.AddField(
            model_name="domain",
            name="tenant",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="domains",
                to="shared_public.Tenant",
            ),
        ),
        migrations.AlterUniqueTogether(name="domain", unique_together={("domain", "folder")}),
    ]



import pytest
from channels.testing import WebsocketCommunicator

from sandbox.routing import domain_application, headers_application


@pytest.mark.asyncio
async def test_on_main_domain():
    communicator = WebsocketCommunicator(
        domain_application,
        "/ws/main/",
        headers=[(b"host", b"localhost")],
    )
    connected, subprotocol = await communicator.connect()

    assert connected

    # Send a message
    await communicator.send_json_to({"message": "hello"})

    # Receive the message
    response = await communicator.receive_json_from()
    assert response["message"] == "www: hello"

    # Close the connection
    await communicator.disconnect()


@pytest.mark.asyncio
async def test_on_tenant_subdomain(tenant3, DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    await DomainModel.objects.acreate(tenant=tenant3, domain="ws_tenant3.localhost")

    communicator = WebsocketCommunicator(
        domain_application,
        "/ws/tenant/",
        headers=[(b"host", b"ws_tenant3.localhost")],
    )
    connected, subprotocol = await communicator.connect()

    assert connected

    # Send a message
    await communicator.send_json_to({"message": "hello"})

    # Receive the message
    response = await communicator.receive_json_from()
    assert response["message"] == "tenant3: hello"

    # Close the connection
    await communicator.disconnect()


@pytest.mark.asyncio
async def test_on_tenant_subfolder(tenant3, DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    await DomainModel.objects.acreate(
        tenant=tenant3, domain="ws_tenants.localhost", folder="tenant3"
    )

    communicator = WebsocketCommunicator(
        domain_application,
        "/tenant3/ws/tenant/",
        headers=[(b"host", b"ws_tenants.localhost")],
    )
    connected, subprotocol = await communicator.connect()

    assert connected

    # Send a message
    await communicator.send_json_to({"message": "hello"})

    # Receive the message
    response = await communicator.receive_json_from()
    assert response["message"] == "tenant3: hello"

    # Close the connection
    await communicator.disconnect()


@pytest.mark.asyncio
async def test_on_main_header():
    communicator = WebsocketCommunicator(
        headers_application,
        "/ws/main/",
        headers=[(b"tenant", b"main")],
    )
    connected, subprotocol = await communicator.connect()

    assert connected

    # Send a message
    await communicator.send_json_to({"message": "hello"})

    # Receive the message
    response = await communicator.receive_json_from()
    assert response["message"] == "www: hello"

    # Close the connection
    await communicator.disconnect()


@pytest.mark.asyncio
async def test_on_tenant_header(TenantModel, db):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    communicator = WebsocketCommunicator(
        headers_application,
        "/ws/tenant/",
        headers=[(b"tenant", b"tenant3")],
    )
    connected, subprotocol = await communicator.connect()

    assert connected

    # Send a message
    await communicator.send_json_to({"message": "hello"})

    # Receive the message
    response = await communicator.receive_json_from()
    assert response["message"] == "tenant3: hello"

    # Close the connection
    await communicator.disconnect()



from django.conf import settings
from django.urls import reverse

from django_pgschemas.routing.info import DomainInfo
from django_pgschemas.routing.urlresolvers import get_dynamic_tenant_prefixed_urlconf


def test_no_tenant():
    url = reverse("profile")

    assert url == "/profile/"


def test_tenant_with_no_folder(tenant1):
    with tenant1:
        url = reverse("profile")

    assert url == "/profile/"


def test_tenant_with_folder(tenant1):
    tenant1.routing = DomainInfo(domain="irrelevant", folder="tenant1")

    dynamic_path = settings.ROOT_URLCONF + "_dynamically_tenant_prefixed"
    urlconf = get_dynamic_tenant_prefixed_urlconf(settings.ROOT_URLCONF, dynamic_path)

    with tenant1:
        url = reverse("profile", urlconf=urlconf)

    tenant1.routing = None

    assert url == "/tenant1/profile/"



from django.contrib import admin
from django.urls import path

from sandbox.views import generic

urlpatterns = [
    path("", generic, name="blog-home"),
    path("entries/", generic, name="entries"),
    path("admin/", admin.site.urls),
]



from django.contrib.auth import get_user_model
from django.db import models


class TenantData(models.Model):
    catalog = models.ForeignKey(
        "shared_public.Catalog", on_delete=models.CASCADE, related_name="tenant_objects"
    )
    user = models.ForeignKey(
        get_user_model(), on_delete=models.CASCADE, related_name="tenant_objects"
    )

    active = models.BooleanField(default=True)



from unittest.mock import MagicMock

import pytest

from django_pgschemas.schema import Schema, activate, deactivate, get_default_schema
from django_pgschemas.signals import schema_activate
from django_pgschemas.utils import schema_exists


def test_schema_activate():
    deactivate()
    schema = Schema.create(schema_name="test")

    receiver = MagicMock()

    schema_activate.connect(receiver)

    activate(schema)

    schema_activate.disconnect(receiver)

    receiver.assert_called_once_with(signal=schema_activate, sender=Schema, schema=schema)


def test_schema_double_activate():
    deactivate()
    schema = Schema.create(schema_name="test")

    receiver = MagicMock()

    schema_activate.connect(receiver)

    activate(schema)
    activate(schema)

    schema_activate.disconnect(receiver)

    receiver.assert_called_once_with(signal=schema_activate, sender=Schema, schema=schema)


def test_schema_deactivate():
    schema = Schema.create(schema_name="test")
    activate(schema)

    receiver = MagicMock()

    schema_activate.connect(receiver)

    deactivate()

    schema_activate.disconnect(receiver)

    receiver.assert_called_once_with(
        signal=schema_activate, sender=Schema, schema=get_default_schema()
    )


def test_schema_override():
    deactivate()
    schema = Schema.create(schema_name="test")

    receiver = MagicMock()

    schema_activate.connect(receiver)

    with schema:
        pass

    schema_activate.disconnect(receiver)

    receiver.assert_called_once_with(signal=schema_activate, sender=Schema, schema=schema)


def test_tenant_delete_callback(TenantModel, db):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")

    backup_create, backup_drop = TenantModel.auto_create_schema, TenantModel.auto_drop_schema
    TenantModel.auto_create_schema = False
    TenantModel.auto_drop_schema = True

    tenant = TenantModel(schema_name="tenant_signal")
    tenant.save()
    tenant.create_schema(sync_schema=False)

    assert schema_exists("tenant_signal")

    TenantModel.objects.all().delete()

    assert not schema_exists("tenant_signal")

    TenantModel.auto_create_schema, TenantModel.auto_drop_schema = backup_create, backup_drop



import os
import shutil
import tempfile

import pytest
from django.core.files.base import ContentFile

from django_pgschemas.contrib.storage import TenantFileSystemStorage
from django_pgschemas.routing.info import DomainInfo
from django_pgschemas.schema import Schema

STORAGE_BASE_URL = "/base-url/"


@pytest.fixture
def temp_dir():
    value = tempfile.mkdtemp()

    yield value

    shutil.rmtree(value)


@pytest.fixture
def storage(temp_dir):
    return TenantFileSystemStorage(location=temp_dir, base_url=STORAGE_BASE_URL)


@pytest.fixture
def settings_pathname(settings):
    settings.PGSCHEMAS_PATHNAME_FUNCTION = lambda tenant: f"custom-pathname-{tenant.schema_name}"


class TestPathIdentifier:
    def test_basic_dynamic(self, storage, tenant1):
        with tenant1:
            assert storage.get_schema_path_identifier() == tenant1.schema_name

    def test_basic_static(self, storage):
        with Schema.create(schema_name="www"):
            assert storage.get_schema_path_identifier() == "www"

    def test_method_in_tenant(self, storage, tenant1):
        tenant1.schema_pathname = lambda: "custom-pathname"
        with tenant1:
            assert storage.get_schema_path_identifier() == "custom-pathname"
        del tenant1.schema_pathname

    def test_function_in_settings(self, tenant1, storage, settings_pathname):
        with tenant1:
            assert storage.get_schema_path_identifier() == f"custom-pathname-{tenant1.schema_name}"


def test_base_location(storage, temp_dir, tenant1, settings_pathname):
    with tenant1:
        assert storage.base_location == f"{temp_dir}/custom-pathname-{tenant1.schema_name}/"


def test_base_url(storage, tenant1):
    tenant1.routing = DomainInfo(domain="irrelevant", folder="tenant1")

    with tenant1:
        assert storage.base_url == STORAGE_BASE_URL

    tenant1.routing = None


def test_file_path(storage, temp_dir, tenant1):
    assert not storage.exists("test.file")

    with tenant1:
        f = ContentFile("random content")
        f_name = storage.save("test.file", f)

        assert os.path.join(temp_dir, tenant1.schema_name, f_name) == storage.path(f_name)

        storage.delete(f_name)

    assert not storage.exists("test.file")


def test_file_save_with_path(storage, temp_dir, tenant1):
    assert not storage.exists("path/to")

    with tenant1:
        storage.save("path/to/test.file", ContentFile("file saved with path"))

        assert storage.exists("path/to")

        with storage.open("path/to/test.file") as f:
            assert f.read() == b"file saved with path"

        assert os.path.exists(
            os.path.join(temp_dir, tenant1.schema_name, "path", "to", "test.file")
        )

        storage.delete("path/to/test.file")

    assert not storage.exists("test.file")


def test_file_url_simple(storage, tenant1):
    tenant1.routing = DomainInfo(domain="irrelevant", folder="tenant1")

    with tenant1:
        assert storage.url("test.file") == "/base-url/test.file"

    tenant1.routing = None


def test_file_url_complex(storage, tenant1):
    with tenant1:
        assert (
            storage.url(r"~!*()'@#$%^&*abc`+ =.file")
            == f"/base-url/{tenant1.schema_name}/~!*()'%40%23%24%25%5E%26*abc%60%2B%20%3D.file"
        )
        assert storage.url("ab\0c") == f"/base-url/{tenant1.schema_name}/ab%00c"
        assert storage.url("a/b\\c.file") == f"/base-url/{tenant1.schema_name}/a/b/c.file"
        assert storage.url("") == f"/base-url/{tenant1.schema_name}/"
        assert storage.url(None) == f"/base-url/{tenant1.schema_name}/"



from django.conf import settings
from django.db import models

from django_pgschemas.models import TenantModel
from django_pgschemas.routing.models import DomainModel


class Tenant(TenantModel):
    pass


if settings.TENANTS.get("default", {}).get("DOMAIN_MODEL", None) is not None:

    class Domain(DomainModel):
        pass


class Catalog(models.Model):
    pass



"""
Django settings for sandbox project.

Generated by 'django-admin startproject' using Django 2.1.4.

For more information on this file, see
https://docs.djangoproject.com/en/2.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.1/ref/settings/
"""

import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "asd#$#ae)^gegm6m9omvic^ct@*@bkf!0afe*+4h$5-zmf^h&$u4(1vr"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = [".localhost"]

TENANTS = {
    "public": {
        "APPS": [
            "sandbox.shared_public",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.staticfiles",
            "django.contrib.messages",
        ],
    },
    "www": {
        "APPS": [
            "sandbox.shared_common",
            "sandbox.app_main",
            "django.contrib.sessions",
            "django.contrib.admin",
        ],
        "URLCONF": "sandbox.app_main.urls",
        "WS_URLCONF": "sandbox.app_main.ws_urls",
        "DOMAINS": ["localhost"],
        "SESSION_KEY": "main",
        "HEADER": "main",
        "FALLBACK_DOMAINS": ["tenants.localhost"],
    },
    "blog": {
        "APPS": [
            "sandbox.shared_common",
            "sandbox.app_blog",
            "django.contrib.sessions",
            "django.contrib.admin",
        ],
        "URLCONF": "sandbox.app_blog.urls",
        "DOMAINS": ["blog.localhost"],
    },
    "default": {
        "TENANT_MODEL": "shared_public.Tenant",
        "DOMAIN_MODEL": "shared_public.Domain",
        "APPS": [
            "sandbox.shared_common",
            "sandbox.app_tenants",
            "django.contrib.sessions",
            "django.contrib.admin",
        ],
        "URLCONF": "sandbox.app_tenants.urls",
        "WS_URLCONF": "sandbox.app_tenants.ws_urls",
        "CLONE_REFERENCE": "sample",
    },
}

# Application definition

INSTALLED_APPS = ["django_pgschemas"]
for schema in TENANTS:
    INSTALLED_APPS += [app for app in TENANTS[schema]["APPS"] if app not in INSTALLED_APPS]

ROOT_URLCONF = TENANTS["default"]["URLCONF"]

ASGI_APPLICATION = "routing.application"

AUTH_USER_MODEL = "shared_common.User"
LOGIN_URL = "login"

MIDDLEWARE = [
    "django_pgschemas.routing.middleware.DomainRoutingMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            f"{BASE_DIR}/sandbox/templates",
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]


# Database
# https://docs.djangoproject.com/en/2.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django_pgschemas.postgresql",
        "NAME": "sandbox",
        "USER": "postgres",
        "PASSWORD": os.environ.get("DATABASE_PASSWORD", "postgres"),
        "HOST": os.environ.get("DATABASE_HOST", "localhost"),
        "PORT": "",
    }
}

DATABASE_ROUTERS = ("django_pgschemas.routers.TenantAppsRouter",)

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "KEY_FUNCTION": "django_pgschemas.contrib.cache.make_key",
    }
}


# Password validation
# https://docs.djangoproject.com/en/2.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]


# Internationalization
# https://docs.djangoproject.com/en/2.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True


USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.1/howto/static-files/

STATIC_URL = "/static/"

DEFAULT_AUTO_FIELD = "django.db.models.AutoField"



from unittest.mock import patch

import pytest
from django.core.management import call_command
from django.core.management.base import CommandError

from django_pgschemas import utils


@pytest.fixture(autouse=True)
def _setup(TenantModel):
    if TenantModel is None:
        pytest.skip("Dynamic tenants are not in use")


def test_cloneschema(transactional_db):
    utils._create_clone_schema_function()

    assert not utils.schema_exists("cloned")

    call_command("cloneschema", "sample", "cloned", verbosity=0)  # All good

    assert utils.schema_exists("cloned")

    with pytest.raises(CommandError):  # Existing destination
        call_command("cloneschema", "sample", "cloned", verbosity=0)

    with pytest.raises(CommandError):  # Not existing source
        call_command("cloneschema", "nonexisting", "newschema", verbosity=0)

    utils.drop_schema("cloned")


def test_createrefschema(transactional_db):
    utils.drop_schema("cloned")
    call_command("createrefschema", verbosity=0)  # All good

    assert utils.schema_exists("sample")

    utils.drop_schema("cloned")
    call_command("createrefschema", recreate=True, verbosity=0)  # All good too

    assert utils.schema_exists("sample")

    utils.drop_schema("cloned")
    call_command("createrefschema", recreate=True, verbosity=0)  # All good too

    assert utils.schema_exists("sample")


def test_interactive_cloneschema(transactional_db):
    answer_provider = (
        n
        for n in [
            "y",  # Would you like to create a database entry?
            "",  # Domain name, simulated wrong answer
            "tenant1copy.localhost",  # Domain name, good answer
        ]
    )

    def patched_input(*args, **kwargs):
        return next(answer_provider)

    with patch("builtins.input", patched_input):
        call_command(
            "cloneschema",
            "tenant1",
            "tenant1copy",
            verbosity=0,
        )

    assert utils.schema_exists("tenant1copy")

    utils.drop_schema("tenant1copy")



from unittest.mock import patch

from django.apps import apps


@patch("django_pgschemas.checks.ensure_tenant_dict")
@patch("django_pgschemas.checks.ensure_public_schema")
@patch("django_pgschemas.checks.ensure_default_schemas")
@patch("django_pgschemas.checks.ensure_overall_schemas")
@patch("django_pgschemas.checks.ensure_extra_search_paths")
def test_all_checkers_called(*checkers):
    config = apps.get_app_config("django_pgschemas")

    config.ready()

    for checker in checkers:
        checker.assert_called()



# Generated by Django 3.0.5 on 2021-02-22 21:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("app_tenants", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="tenantdata",
            name="active",
            field=models.BooleanField(default=True),
        ),
    ]



from django.contrib.auth import get_user_model
from django.db import models


class BlogEntry(models.Model):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE, related_name="blogs")



from itertools import permutations
from unittest.mock import MagicMock

import pytest
from django.http import Http404

from django_pgschemas.routing.info import DomainInfo, HeadersInfo, SessionInfo
from django_pgschemas.routing.middleware import (
    DomainRoutingMiddleware,
    HeadersRoutingMiddleware,
    SessionRoutingMiddleware,
    strip_tenant_from_path_factory,
)


@pytest.mark.parametrize(
    "path, prefix, expected",
    [
        ("/some/path/", "", "/some/path/"),
        ("/some/path/", "path", "/some/path/"),
        ("/some/path/", "some", "/path/"),
    ],
)
def test_strip_tenant_from_path_factory(path, prefix, expected):
    actual = strip_tenant_from_path_factory(prefix)(path)

    assert actual == expected


class FakeRequest:
    def __init__(
        self,
        *,
        domain: str = "",
        path: str = "",
        session_tenant_ref: str | None = None,
        headers_tenant_ref: str | None = None,
    ) -> None:
        self.domain = domain
        self.path = path
        self.session_tenant_ref = session_tenant_ref
        self.headers_tenant_ref = headers_tenant_ref

    def get_host(self) -> str:
        return self.domain

    @property
    def session(self) -> dict:
        return {
            "tenant": self.session_tenant_ref,
        }

    @property
    def headers(self) -> dict:
        return {
            "tenant": self.headers_tenant_ref,
        }


class TestDomainRoutingMiddleware:
    @pytest.fixture(autouse=True)
    def _setup(self, tenant1, tenant2, DomainModel):
        if DomainModel is None:
            pytest.skip("Domain model is not in use")

        DomainModel.objects.create(
            tenant=tenant1,
            domain="tenant1.localhost",
            is_primary=True,
        )
        DomainModel.objects.create(
            tenant=tenant1,
            domain="tenants.localhost",
            folder="tenant1",
            is_primary=False,
        )
        DomainModel.objects.create(
            tenant=tenant2,
            domain="tenant2.localhost",
            is_primary=True,
        )
        DomainModel.objects.create(
            tenant=tenant2,
            domain="tenants.localhost",
            folder="tenant2",
            is_primary=False,
        )

    @pytest.mark.parametrize(
        "domain, path, schema_name",
        [
            ("tenant1.localhost", "", "tenant1"),
            ("tenants.localhost", "tenant1", "tenant1"),
            ("tenant2.localhost", "", "tenant2"),
            ("tenants.localhost", "tenant2", "tenant2"),
            ("tenant3.localhost", "", None),
            ("localhost", "", "www"),
            ("blog.localhost", "", "blog"),
            ("tenants.localhost", "", "www"),  # fallback domains
        ],
    )
    def test_tenant_matching(self, domain, path, schema_name, db):
        request = FakeRequest(domain=domain, path=f"/{path}/some/path/")
        get_response = MagicMock()

        handler = DomainRoutingMiddleware(get_response)

        if schema_name is None:
            with pytest.raises(Http404):
                handler(request)
        else:
            handler(request)
            assert request.tenant is not None
            assert request.tenant.schema_name == schema_name
            assert isinstance(request.tenant.routing, DomainInfo)
            assert request.tenant.routing.domain == domain
            assert request.tenant.routing.folder == (path if path else None)


class TestDomainRoutingMiddlewareRedirection:
    @pytest.fixture(autouse=True)
    def _setup(self, tenant1, tenant2, DomainModel):
        if DomainModel is None:
            pytest.skip("Domain model is not in use")

        DomainModel(domain="tenant1.localhost", tenant=tenant1).save()
        DomainModel(
            domain="tenant1redirect.localhost",
            tenant=tenant1,
            is_primary=False,
            redirect_to_primary=True,
        ).save()
        DomainModel(
            domain="everyone.localhost",
            folder="tenant1redirect",
            tenant=tenant1,
            is_primary=False,
            redirect_to_primary=True,
        ).save()

        DomainModel(domain="everyone.localhost", folder="tenant2", tenant=tenant2).save()
        DomainModel(
            domain="tenant2redirect.localhost",
            tenant=tenant2,
            is_primary=False,
            redirect_to_primary=True,
        ).save()
        DomainModel(
            domain="everyone.localhost",
            folder="tenant2redirect",
            tenant=tenant2,
            is_primary=False,
            redirect_to_primary=True,
        ).save()

    @pytest.mark.parametrize(
        "domain, path, expected_redirection",
        [
            (
                "tenant1redirect.localhost",
                "/some/random/url/",
                "//tenant1.localhost/some/random/url/",
            ),
            (
                "everyone.localhost",
                "/tenant1redirect/some/random/url/",
                "//tenant1.localhost/some/random/url/",
            ),
            (
                "tenant2redirect.localhost",
                "/some/random/url/",
                "//everyone.localhost/tenant2/some/random/url/",
            ),
            (
                "everyone.localhost",
                "/tenant2redirect/some/random/url/",
                "//everyone.localhost/tenant2/some/random/url/",
            ),
        ],
    )
    def test_redirection(self, domain, path, expected_redirection):
        request = FakeRequest(domain=domain, path=path)
        get_response = MagicMock()

        response = DomainRoutingMiddleware(get_response)(request)

        assert response.status_code == 301
        assert response.url == expected_redirection
        assert response["Location"] == expected_redirection


class TestSessionRoutingMiddleware:
    @pytest.mark.parametrize(
        "session_key, schema_name",
        [
            ("www", "www"),
            ("main", "www"),
            ("blog", "blog"),
            ("tenant1", "tenant1"),
        ],
    )
    def test_tenant_matching(self, DomainModel, session_key, schema_name, db):
        if DomainModel is None and "tenant" in schema_name:
            pytest.skip("Domain model is not in use")

        request = FakeRequest(session_tenant_ref=session_key)
        get_response = MagicMock()

        handler = SessionRoutingMiddleware(get_response)

        handler(request)

        assert request.tenant is not None
        assert request.tenant.schema_name == schema_name
        assert isinstance(request.tenant.routing, SessionInfo)
        assert request.tenant.routing.reference == session_key


class TestHeadersRoutingMiddleware:
    @pytest.mark.parametrize(
        "header, schema_name",
        [
            ("www", "www"),
            ("main", "www"),
            ("blog", "blog"),
            ("tenant1", "tenant1"),
        ],
    )
    def test_tenant_matching(self, DomainModel, header, schema_name, db):
        if DomainModel is None and "tenant" in schema_name:
            pytest.skip("Domain model is not in use")

        request = FakeRequest(headers_tenant_ref=header)
        get_response = MagicMock()

        handler = HeadersRoutingMiddleware(get_response)

        handler(request)

        assert request.tenant is not None
        assert request.tenant.schema_name == schema_name
        assert isinstance(request.tenant.routing, HeadersInfo)
        assert request.tenant.routing.reference == header


@pytest.mark.parametrize(
    "first_middleware, second_middleware, last_middleware",
    permutations([DomainRoutingMiddleware, SessionRoutingMiddleware, HeadersRoutingMiddleware]),
)
def test_last_middleware_prevails(
    first_middleware, second_middleware, last_middleware, tenant1, tenant2, tenant3, DomainModel
):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    DomainModel.objects.create(
        domain="tenants.localhost",
        tenant=tenant1,
        folder="tenant1",
        is_primary=True,
    )

    request = FakeRequest(
        domain="tenants.localhost",
        path="/tenant1/some/path/",
        session_tenant_ref=tenant2.schema_name,
        headers_tenant_ref=tenant3.schema_name,
    )
    get_response = MagicMock()

    handler = first_middleware(second_middleware(last_middleware(get_response)))

    handler(request)

    if last_middleware is DomainRoutingMiddleware:
        assert request.tenant == tenant1
        assert isinstance(request.tenant.routing, DomainInfo)

    if last_middleware is SessionRoutingMiddleware:
        assert request.tenant == tenant2
        assert isinstance(request.tenant.routing, SessionInfo)

    if last_middleware is HeadersRoutingMiddleware:
        assert request.tenant == tenant3
        assert isinstance(request.tenant.routing, HeadersInfo)



# Generated by Django 2.1.4 on 2019-01-13 03:07

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(blank=True, null=True, verbose_name="last login"),
                ),
                ("email", models.EmailField(max_length=254, unique=True)),
                ("display_name", models.CharField(max_length=50)),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.conf import settings
from django.urls import reverse

from django_pgschemas.management.commands import SchemaCommand
from django_pgschemas.routing.info import DomainInfo
from django_pgschemas.routing.models import get_primary_domain_for_tenant
from django_pgschemas.routing.urlresolvers import get_urlconf_from_schema
from django_pgschemas.schema import Schema


class Command(SchemaCommand):
    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            dest="url_name",
            help="Url name to resolve in the specified schema",
        )

    def handle_schema(self, schema: Schema, *args, **options):
        if schema.is_dynamic:
            primary_domain = get_primary_domain_for_tenant(schema)
            schema.routing = DomainInfo(domain=primary_domain.domain, folder=primary_domain.folder)
        else:
            schema.routing = DomainInfo(domain=settings.TENANTS[schema.schema_name]["DOMAINS"][0])
        self.stdout.write(reverse(options["url_name"], urlconf=get_urlconf_from_schema(schema)))



from django.contrib.auth.models import AbstractBaseUser, BaseUserManager
from django.db import models


class UserManager(BaseUserManager):
    pass


class User(AbstractBaseUser):
    email = models.EmailField(unique=True)
    display_name = models.CharField(max_length=50)

    USERNAME_FIELD = "email"
    REQUIRED_FIELDS = ("display_name",)

    objects = UserManager()



import pytest
from django.apps import apps
from django.core.checks import Critical, Error, Warning
from django.core.exceptions import ImproperlyConfigured

from django_pgschemas import checks


@pytest.fixture
def app_config():
    return apps.get_app_config("django_pgschemas")


@pytest.fixture
def tenant_manager(TenantModel, db):
    if TenantModel is None:
        yield None
    else:
        backup = TenantModel.auto_create_schema
        TenantModel.auto_create_schema = False

        yield TenantModel._default_manager

        TenantModel.auto_create_schema = backup


def test_get_tenant_app(tenants_settings, TenantModel):
    if TenantModel:
        assert checks.get_tenant_app() == "sandbox.shared_public"

        del tenants_settings["default"]

    assert checks.get_tenant_app() is None


def test_get_domain_app(tenants_settings, DomainModel):
    if DomainModel:
        assert checks.get_domain_app() == "sandbox.shared_public"

        del tenants_settings["default"]["DOMAIN_MODEL"]

    assert checks.get_domain_app() is None


def test_get_user_app(settings):
    assert checks.get_user_app() == "sandbox.shared_common"

    del settings.AUTH_USER_MODEL

    assert checks.get_user_app() is None


def test_get_session_app(settings):
    assert checks.get_session_app() == "django.contrib.sessions"


def test_ensure_tenant_dict(settings):
    backup = settings.TENANTS

    del settings.TENANTS

    with pytest.raises(ImproperlyConfigured) as ctx:
        checks.ensure_tenant_dict()

    assert str(ctx.value) == "TENANTS dict setting not set."

    settings.TENANTS = backup


class TestEnsurePublicSchema:
    @pytest.mark.parametrize("value", [None, "", 1])
    def test_no_dict(self, tenants_settings, value):
        tenants_settings["public"] = value

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_public_schema()

        assert str(ctx.value) == "TENANTS must contain a 'public' dict."

    @pytest.mark.parametrize(
        "member",
        [
            "URLCONF",
            "WS_URLCONF",
            "DOMAINS",
            "FALLBACK_DOMAINS",
            "SESSION_KEY",
            "HEADER",
        ],
    )
    def test_invalid_members(self, tenants_settings, member):
        tenants_settings["public"][member] = None

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_public_schema()

        assert str(ctx.value) == f"TENANTS['public'] cannot contain a '{member}' key."


class TestEnsureDefaultSchema:
    def test_static_only(self, tenants_settings):
        if "default" in tenants_settings:
            del tenants_settings["default"]

        checks.ensure_default_schemas()

    @pytest.mark.parametrize("value", [None, "", 1])
    def test_no_dict(self, tenants_settings, value):
        tenants_settings["default"] = value

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_default_schemas()

        assert str(ctx.value) == "TENANTS must contain a 'default' dict."

    @pytest.mark.parametrize(
        "member",
        [
            "DOMAINS",
            "FALLBACK_DOMAINS",
            "SESSION_KEY",
            "HEADER",
        ],
    )
    def test_invalid_members(self, tenants_settings, member):
        if "default" not in tenants_settings:
            pytest.skip("default not in tenant settings")

        tenants_settings["default"][member] = None

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_default_schemas()

        assert str(ctx.value) == f"TENANTS['default'] cannot contain a '{member}' key."

    @pytest.mark.parametrize(
        "member",
        [
            "TENANT_MODEL",
            "URLCONF",
        ],
    )
    def test_required_members(self, tenants_settings, member):
        if "default" not in tenants_settings:
            pytest.skip("default not in tenant settings")

        del tenants_settings["default"][member]

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_default_schemas()

        assert str(ctx.value) == f"TENANTS['default'] must contain a '{member}' key."

    @pytest.mark.parametrize(
        "name",
        [
            "public",
            "default",
            "blog",
            "www",
        ],
    )
    def test_clone_reference_invalid_name(self, tenants_settings, name):
        if "default" not in tenants_settings:
            pytest.skip("default not in tenant settings")

        tenants_settings["default"]["CLONE_REFERENCE"] = name

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_default_schemas()

        assert (
            str(ctx.value) == "TENANTS['default']['CLONE_REFERENCE'] must be a unique schema name."
        )


class TestEnsureOverallSchema:
    @pytest.mark.parametrize(
        "name",
        [
            "pg_something",
            "1something",
            ".something",
            "-something",
            "tomanycharacters0123456789abcdef0123456789abcef0123456789abcdef0",
        ],
    )
    def test_invalid_names(self, tenants_settings, name):
        tenants_settings[name] = {}

        with pytest.raises(ImproperlyConfigured) as ctx:
            checks.ensure_overall_schemas()

        assert str(ctx.value) == f"'{name}' is not a valid schema name."


@pytest.mark.parametrize(
    "extra",
    [
        "public",
        "blog",
        "www",
        # "default",
    ],
)
def test_ensure_extra_search_paths(settings, extra, db):
    settings.PGSCHEMAS_EXTRA_SEARCH_PATHS = [extra]

    with pytest.raises(ImproperlyConfigured) as ctx:
        checks.ensure_extra_search_paths()

    invalid = ", ".join([extra])
    assert str(ctx.value) == f"Do not include '{invalid}' on PGSCHEMAS_EXTRA_SEARCH_PATHS."


class TestCheckPrincipalApps:
    BASE_DEFAULT = {"TENANT_MODEL": "shared_public.Tenant", "DOMAIN_MODEL": "shared_public.DOMAIN"}

    def test_location_wrong(self, tenants_settings, app_config):
        tenants_settings.update(
            {
                "public": {"APPS": []},
                "default": self.BASE_DEFAULT,
            }
        )
        expected_errors = [
            Error(
                "Your tenant app 'sandbox.shared_public' must be on the 'public' schema.",
                id="pgschemas.W001",
            ),
            Error(
                "Your domain app 'sandbox.shared_public' must be on the 'public' schema.",
                id="pgschemas.W001",
            ),
        ]

        errors = checks.check_principal_apps(app_config)
        assert errors == expected_errors

    def test_location_twice(self, tenants_settings, app_config):
        tenants_settings.update(
            {
                "public": {"APPS": ["sandbox.shared_public"]},
                "default": {**self.BASE_DEFAULT, "APPS": ["sandbox.shared_public"]},
            }
        )

        expected_errors = [
            Error(
                "Your tenant app 'sandbox.shared_public' in TENANTS['default']['APPS'] "
                "must be on the 'public' schema only.",
                id="pgschemas.W001",
            ),
            Error(
                "Your domain app 'sandbox.shared_public' in TENANTS['default']['APPS'] "
                "must be on the 'public' schema only.",
                id="pgschemas.W001",
            ),
        ]

        errors = checks.check_principal_apps(app_config)
        assert errors == expected_errors


class TestCheckOtherApps:
    def test_contenttypes_location_wrong(self, tenants_settings, app_config):
        tenants_settings.update(
            {
                "default": {"APPS": ["django.contrib.contenttypes"]},
            }
        )
        expected_errors = [
            Warning(
                "'django.contrib.contenttypes' in TENANTS['default']['APPS'] "
                "must be on 'public' schema only.",
                id="pgschemas.W002",
            )
        ]

        errors = checks.check_other_apps(app_config)
        assert errors == expected_errors

    def test_contenttypes_location_twice(self, tenants_settings, app_config):
        tenants_settings.update(
            {
                "default": {},
                "www": {"APPS": ["django.contrib.contenttypes"]},
            }
        )
        expected_errors = [
            Warning(
                "'django.contrib.contenttypes' in TENANTS['www']['APPS'] "
                "must be on 'public' schema only.",
                id="pgschemas.W002",
            )
        ]

        errors = checks.check_other_apps(app_config)
        assert errors == expected_errors

    def test_user_location_wrong(self, tenants_settings, app_config):
        user_app = checks.get_user_app()
        tenants_settings.update(
            {
                "default": {"APPS": ["django.contrib.sessions"]},
            }
        )
        expected_errors = [
            Warning(
                f"'{user_app}' must be together with 'django.contrib.sessions' "
                "in TENANTS['default']['APPS'].",
                id="pgschemas.W003",
            )
        ]

        errors = checks.check_other_apps(app_config)
        assert errors == expected_errors

    def test_session_location_wrong(self, tenants_settings, app_config):
        user_app = checks.get_user_app()
        tenants_settings.update(
            {
                "www": {"APPS": ["shared_common", user_app]},
                "default": {"APPS": ["shared_common"]},
            }
        )
        expected_errors = [
            Warning(
                f"'django.contrib.sessions' must be together with '{user_app}' "
                "in TENANTS['www']['APPS'].",
                id="pgschemas.W003",
            )
        ]

        errors = checks.check_other_apps(app_config)
        assert errors == expected_errors


@pytest.mark.parametrize("schema", ["public", "www", "blog", "sample"])
def test_check_schema_names(schema, app_config, tenant_manager):
    if tenant_manager is None:
        pytest.skip("Dynamic tenants are not in use")

    tenant_manager.create(schema_name=schema)
    expected_errors = [
        Critical(
            f"Name clash found between static and dynamic tenants: {{'{schema}'}}",
            id="pgschemas.W004",
        ),
    ]

    errors = checks.check_schema_names(app_config)
    assert errors == expected_errors



import pytest
from django.core import management
from django.urls.exceptions import NoReverseMatch


def test_urls_for_main_error(db):
    with pytest.raises(NoReverseMatch):
        management.call_command("reverse_url", "entries", schemas=["www"])


def test_urls_for_main_success(stdout, db):
    management.call_command("reverse_url", "register", schemas=["www"], stdout=stdout)
    stdout.seek(0)
    assert stdout.read().strip() == "/register/"


def test_urls_for_blog_error(db):
    with pytest.raises(NoReverseMatch):
        management.call_command("reverse_url", "register", schemas=["blog"])


def test_urls_for_blog_success(stdout, db):
    management.call_command("reverse_url", "entries", schemas=["blog"], stdout=stdout)
    stdout.seek(0)
    assert stdout.read().strip() == "/entries/"



from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter

from django_pgschemas.contrib.channels import (
    DomainRoutingMiddleware,
    HeadersRoutingMiddleware,
    TenantURLRouter,
)

domain_application = ProtocolTypeRouter(
    {
        "websocket": DomainRoutingMiddleware(
            AuthMiddlewareStack(
                TenantURLRouter(),
            ),
        ),
    }
)

headers_application = ProtocolTypeRouter(
    {
        "websocket": HeadersRoutingMiddleware(
            AuthMiddlewareStack(
                TenantURLRouter(),
            ),
        ),
    }
)

application = domain_application



import pytest

from django_pgschemas.routing.models import get_primary_domain_for_tenant


@pytest.fixture(autouse=True)
def _setup(DomainModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")


@pytest.mark.parametrize(
    "domain, folder, expected",
    [
        ("tenants.localhost", "", "tenants.localhost"),
        ("tenants.localhost", "tenant1", "tenants.localhost/tenant1"),
    ],
)
def test_str(tenant1, domain, folder, expected, DomainModel):
    item = DomainModel.objects.create(
        tenant=tenant1,
        domain=domain,
        folder=folder,
    )

    assert str(item) == expected


def test_only_one_primary(tenant1, DomainModel):
    domain1 = DomainModel.objects.create(
        tenant=tenant1,
        domain="tenant1.localhost",
        folder="",
        is_primary=True,
    )

    assert domain1.is_primary

    domain2 = DomainModel.objects.create(
        tenant=tenant1,
        domain="tenants.localhost",
        folder="tenant1",
        is_primary=True,
    )

    domain1.refresh_from_db()

    assert not domain1.is_primary
    assert domain2.is_primary


@pytest.mark.parametrize("is_primary", [True, False])
def test_redirect_to_primary_if_primary(tenant1, is_primary, DomainModel):
    domain1 = DomainModel.objects.create(
        domain="tenant1.localhost",
        folder="",
        is_primary=is_primary,
        redirect_to_primary=True,
        tenant=tenant1,
    )

    assert domain1.redirect_to_primary is not domain1.is_primary


@pytest.mark.parametrize(
    "domain, folder, path, expected",
    [
        ("tenants.localhost", "", "", "//tenants.localhost/"),
        ("tenants.localhost", "tenant1", "", "//tenants.localhost/tenant1/"),
        ("tenants.localhost", "tenant1", "some/path", "//tenants.localhost/tenant1/some/path"),
        ("tenants.localhost", "tenant1", "/some/path", "//tenants.localhost/tenant1/some/path"),
        ("tenants.localhost", "tenant1", "/some/path/", "//tenants.localhost/tenant1/some/path/"),
    ],
)
def test_absolute_url(tenant1, domain, folder, path, expected, DomainModel):
    item = DomainModel.objects.create(
        tenant=tenant1,
        domain=domain,
        folder=folder,
    )

    assert item.absolute_url(path) == expected


@pytest.mark.parametrize("is_primary", [True, False, None])
def test_get_primary_domain_for_tenant(tenant1, is_primary, DomainModel):
    if is_primary is not None:
        item = DomainModel.objects.create(
            tenant=tenant1,
            domain="tenant1.localhost",
        )
        DomainModel.objects.update(is_primary=is_primary)
        item.refresh_from_db()

    if is_primary:
        assert get_primary_domain_for_tenant(tenant1) == item
    else:
        assert get_primary_domain_for_tenant(tenant1) is None



from django_pgschemas.log import SchemaContextFilter
from django_pgschemas.routing.info import DomainInfo, HeadersInfo, SessionInfo
from django_pgschemas.schema import Schema


class FakeRecord:
    pass


class TestSchemaContextFilter:
    def test_filter_with_domain(self):
        record = FakeRecord()
        scf = SchemaContextFilter()

        with Schema.create(
            schema_name="some-tenant",
            routing=DomainInfo(
                domain="some-tenant.some-url.com",
                folder="folder1",
            ),
        ):
            scf.filter(record)

        assert record.schema_name == "some-tenant"
        assert record.domain == "some-tenant.some-url.com"
        assert record.folder == "folder1"

    def test_filter_with_session(self):
        record = FakeRecord()
        scf = SchemaContextFilter()

        with Schema.create(
            schema_name="some-tenant",
            routing=SessionInfo(reference="tenant1"),
        ):
            scf.filter(record)

        assert record.schema_name == "some-tenant"
        assert record.reference == "tenant1"

    def test_filter_with_header(self):
        record = FakeRecord()
        scf = SchemaContextFilter()

        with Schema.create(
            schema_name="some-tenant",
            routing=HeadersInfo(reference="tenant1"),
        ):
            scf.filter(record)

        assert record.schema_name == "some-tenant"
        assert record.reference == "tenant1"



# Generated by Django 3.2.4 on 2021-10-27 19:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("shared_public", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="domain",
            name="redirect_to_primary",
            field=models.BooleanField(default=False),
        ),
    ]



import json

from channels.generic.websocket import AsyncWebsocketConsumer


class EchoConsumer(AsyncWebsocketConsumer):
    async def connect(self):
        await self.accept()

    async def disconnect(self, close_code):
        pass

    async def receive(self, text_data):
        data = json.loads(text_data)
        message = data["message"]

        schema_name = self.scope["tenant"].schema_name
        await self.send(text_data=json.dumps({"message": f"{schema_name}: {message}"}))



from contextlib import contextmanager

import pytest
from django.apps import apps
from django.contrib.auth import authenticate
from django.db import ProgrammingError, transaction

from django_pgschemas.schema import Schema, deactivate
from django_pgschemas.signals import dynamic_tenant_post_sync
from django_pgschemas.utils import schema_exists


@pytest.fixture
def BlogEntryModel():
    return apps.get_model("app_blog.BlogEntry")


@pytest.fixture
def CatalogModel():
    return apps.get_model("shared_public.Catalog")


@pytest.fixture
def MainDataModel():
    return apps.get_model("app_main.MainData")


@pytest.fixture
def UserModel():
    return apps.get_model("shared_common.User")


@pytest.fixture
def TenantDataModel(TenantModel):
    return apps.get_model("app_tenants.TenantData") if TenantModel is not None else None


class ControlledException(Exception):
    pass


@contextmanager
def controlled_raises(exception):
    """
    Since we are expecting database errors, we must use savepoints in order
    to make sure multiple errors can be caught in the same test case.
    """
    sid = transaction.savepoint()

    with pytest.raises(exception):
        yield

    transaction.savepoint_rollback(sid)


class TestTenantAutomaticOperations:
    @pytest.fixture(autouse=True)
    def _setup(self, TenantModel):
        if TenantModel is None:
            pytest.skip("Dynamic tenants are not in use")

    def test_new_creation_deletion(self, TenantModel, db):
        assert not schema_exists("new_tenant1")

        tenant = TenantModel(schema_name="new_tenant1")
        tenant.save(verbosity=0)

        assert schema_exists("new_tenant1")

        tenant.delete(force_drop=True)

        assert not schema_exists("new_tenant1")

    def test_existing_creation(self, TenantModel, db):
        assert not schema_exists("new_tenant1")

        tenant = TenantModel(schema_name="new_tenant1")
        tenant.auto_create_schema = False
        tenant.save(verbosity=0)

        assert not schema_exists("new_tenant1")

        tenant.auto_create_schema = True
        tenant.save(verbosity=0)

        assert schema_exists("new_tenant1")

        tenant.delete(force_drop=True)

        assert not schema_exists("new_tenant1")

    def test_new_aborted_creation(self, TenantModel, db):
        def signal_receiver(*args, **kwargs):
            raise ControlledException

        assert not schema_exists("new_tenant1")

        tenant = TenantModel(schema_name="new_tenant1")
        dynamic_tenant_post_sync.connect(signal_receiver)

        with pytest.raises(ControlledException):
            tenant.save(verbosity=0)

        assert not schema_exists("new_tenant1")
        assert not TenantModel.objects.filter(schema_name="new_tenant1").exists()

        dynamic_tenant_post_sync.disconnect(signal_receiver)

    def test_existing_aborted_creation(self, TenantModel, db):
        def signal_receiver(*args, **kwargs):
            raise ControlledException

        assert not schema_exists("new_tenant1")

        tenant = TenantModel(schema_name="new_tenant1")
        tenant.auto_create_schema = False
        tenant.save(verbosity=0)

        tenant.auto_create_schema = True
        dynamic_tenant_post_sync.connect(signal_receiver)

        with pytest.raises(ControlledException):
            tenant.save(verbosity=0)

        assert not schema_exists("new_tenant1")
        assert TenantModel.objects.filter(schema_name="new_tenant1").exists()

        dynamic_tenant_post_sync.disconnect(signal_receiver)

        tenant.delete(force_drop=True)

        assert not TenantModel.objects.filter(schema_name="new_tenant1").exists()


class TestTenantIntegration:
    @pytest.fixture(autouse=True)
    def _setup(
        self, tenant1, CatalogModel, UserModel, MainDataModel, BlogEntryModel, TenantDataModel
    ):
        catalog = CatalogModel.objects.create()
        CatalogModel.objects.create()

        with Schema.create(schema_name="www"):
            user = UserModel.objects.create(email="main@localhost", display_name="Main User")
            user.set_password("weakpassword")
            user.save()
            MainDataModel.objects.create()

        with Schema.create(schema_name="blog"):
            user = UserModel.objects.create(email="blog@localhost", display_name="Blog User")
            user.set_password("weakpassword")
            user.save()
            BlogEntryModel.objects.create(user=user)

        if TenantDataModel is not None:
            with tenant1:
                user = UserModel.objects.create(
                    email="tenant@localhost", display_name="Tenant User"
                )
                user.set_password("weakpassword")
                user.save()
                TenantDataModel.objects.create(user=user, catalog=catalog)

    def test_migrated_public_apps(
        self, CatalogModel, UserModel, MainDataModel, BlogEntryModel, TenantDataModel
    ):
        deactivate()
        # Apps expected to be migrated

        assert CatalogModel.objects.count() == 2

        # Apps expected to NOT be migrated

        with controlled_raises(ProgrammingError):
            list(UserModel.objects.all())

        with controlled_raises(ProgrammingError):
            list(MainDataModel.objects.all())

        with controlled_raises(ProgrammingError):
            list(BlogEntryModel.objects.all())

        if TenantDataModel is not None:
            with controlled_raises(ProgrammingError):
                list(TenantDataModel.objects.all())

    def test_migrated_main_apps(
        self, CatalogModel, UserModel, MainDataModel, BlogEntryModel, TenantDataModel
    ):
        with Schema.create(schema_name="www"):
            # Apps expected to be migrated

            assert CatalogModel.objects.count() == 2
            assert UserModel.objects.count() == 1
            assert MainDataModel.objects.count() == 1

            # Apps expected to NOT be migrated

            with controlled_raises(ProgrammingError):
                list(BlogEntryModel.objects.all())

            if TenantDataModel is not None:
                with controlled_raises(ProgrammingError):
                    list(TenantDataModel.objects.all())

    def test_migrated_blog_apps(
        self, CatalogModel, UserModel, MainDataModel, BlogEntryModel, TenantDataModel
    ):
        with Schema.create(schema_name="blog"):
            # Apps expected to be migrated

            assert CatalogModel.objects.count() == 2
            assert UserModel.objects.count() == 1
            assert BlogEntryModel.objects.count() == 1

            # Direct and reverse relations
            assert UserModel.objects.first() == BlogEntryModel.objects.first().user
            assert UserModel.objects.first().blogs.first() == BlogEntryModel.objects.first()

            # Apps expected to NOT be migrated

            with controlled_raises(ProgrammingError):
                list(MainDataModel.objects.all())

            if TenantDataModel is not None:
                with controlled_raises(ProgrammingError):
                    list(TenantDataModel.objects.all())

    def test_migrated_tenant_apps(
        self, tenant1, CatalogModel, UserModel, MainDataModel, BlogEntryModel, TenantDataModel
    ):
        if not TenantDataModel:
            pytest.skip("Dynamic tenants are not in use")

        with tenant1:
            # Apps expected to be migrated

            assert CatalogModel.objects.count() == 2
            assert UserModel.objects.count() == 1
            assert TenantDataModel.objects.count() == 1

            # Direct and reverse relations
            assert UserModel.objects.first() == TenantDataModel.objects.first().user
            assert (
                UserModel.objects.first().tenant_objects.first() == TenantDataModel.objects.first()
            )
            assert CatalogModel.objects.first() == TenantDataModel.objects.first().catalog
            assert (
                CatalogModel.objects.first().tenant_objects.first()
                == TenantDataModel.objects.first()
            )

            # Apps expected to NOT be migrated

            with controlled_raises(ProgrammingError):
                list(MainDataModel.objects.all())

            with controlled_raises(ProgrammingError):
                list(BlogEntryModel.objects.all())

    def test_cross_authentication(self, tenant1, TenantModel):
        with Schema.create(schema_name="www"):
            assert authenticate(email="main@localhost", password="weakpassword")
            assert not authenticate(email="blog@localhost", password="weakpassword")
            assert not authenticate(email="tenant@localhost", password="weakpassword")

        with Schema.create(schema_name="blog"):
            assert not authenticate(email="main@localhost", password="weakpassword")
            assert authenticate(email="blog@localhost", password="weakpassword")
            assert not authenticate(email="tenant@localhost", password="weakpassword")

        if TenantModel is not None:
            with tenant1:
                assert not authenticate(email="main@localhost", password="weakpassword")
                assert not authenticate(email="blog@localhost", password="weakpassword")
                assert authenticate(email="tenant@localhost", password="weakpassword")

        with controlled_raises(ProgrammingError):
            authenticate(email="irrelevant@localhost", password="irrelevant")



import pytest
from django.apps import apps
from django.test import Client


@pytest.fixture
def UserModel():
    return apps.get_model("shared_common.User")


@pytest.fixture(autouse=True)
def _setup(tenant1, tenant2, DomainModel, UserModel):
    if DomainModel is None:
        pytest.skip("Domain model is not in use")

    DomainModel.objects.create(
        tenant=tenant1, domain="everyone.localhost", folder="tenant1", is_primary=True
    )
    DomainModel.objects.create(
        tenant=tenant2, domain="everyone.localhost", folder="tenant2", is_primary=True
    )
    with tenant1:
        UserModel.objects.create(email="user1@localhost", display_name="Admin")

    with tenant2:
        UserModel.objects.create(email="user2@localhost", display_name="Admin")


@pytest.fixture
def client():
    return Client(headers={"host": "everyone.localhost"})


@pytest.mark.bug
def test_bug_in_cached_urls_1(client):
    # Provoke redirect to login on tenant2
    client.get("/tenant2/profile/advanced/")

    # Provoke redirect to login on tenant1
    buggy_response = client.get("/tenant1/profile/advanced/")

    assert buggy_response.status_code == 302
    assert buggy_response.url == "/tenant1/login/?next=/tenant1/profile/advanced/"


@pytest.mark.bug
def test_bug_in_cached_urls_2(client):
    # Provoke redirect to login on tenant1
    client.get("/tenant1/profile/advanced/")

    # Provoke redirect to login on tenant2
    buggy_response = client.get("/tenant2/profile/advanced/")

    assert buggy_response.status_code == 302
    assert buggy_response.url == "/tenant2/login/?next=/tenant2/profile/advanced/"
